@@ -1,32 +1,41 @@
 # Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
-# For details: https://bitbucket.org/ned/coveragepy/src/default/NOTICE.txt
+# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt
 
 """Command-line support for coverage.py."""
 
-from __future__ import print_function
 
 import glob
-import optparse
+import optparse     # pylint: disable=deprecated-module
+import os
 import os.path
+import shlex
 import sys
 import textwrap
 import traceback
 
+import coverage
+from coverage import Coverage
 from coverage import env
 from coverage.collector import CTracer
-from coverage.debug import info_formatter, info_header
-from coverage.execfile import run_python_file, run_python_module
-from coverage.misc import BaseCoverageException, ExceptionDuringRun, NoSource
-from coverage.results import should_fail_under
+from coverage.data import line_counts
+from coverage.debug import info_formatter, info_header, short_stack
+from coverage.exceptions import BaseCoverageException, ExceptionDuringRun, NoSource
+from coverage.execfile import PyRunner
+from coverage.misc import human_sorted
+from coverage.results import Numbers, should_fail_under
 
 
-class Opts(object):
+class Opts:
     """A namespace class for individual options we'll build parsers from."""
 
     append = optparse.make_option(
         '-a', '--append', action='store_true',
         help="Append coverage data to .coverage, otherwise it starts clean each time.",
     )
+    keep = optparse.make_option(
+        '', '--keep', action='store_true',
+        help="Keep original coverage files, otherwise they are deleted.",
+    )
     branch = optparse.make_option(
         '', '--branch', action='store_true',
         help="Measure branch coverage in addition to statement coverage.",
@@ -38,13 +47,25 @@ class Opts(object):
         '', '--concurrency', action='store', metavar="LIB",
         choices=CONCURRENCY_CHOICES,
         help=(
-            "Properly measure code using a concurrency library. "
-            "Valid values are: %s."
-        ) % ", ".join(CONCURRENCY_CHOICES),
+            "Properly measure code using a concurrency library. " +
+            "Valid values are: {}."
+        ).format(", ".join(CONCURRENCY_CHOICES)),
+    )
+    context = optparse.make_option(
+        '', '--context', action='store', metavar="LABEL",
+        help="The context label to record for this coverage run.",
+    )
+    contexts = optparse.make_option(
+        '', '--contexts', action='store',
+        metavar="REGEX1,REGEX2,...",
+        help=(
+            "Only display data from lines covered in the given contexts. " +
+            "Accepts Python regexes, which must be quoted."
+        ),
     )
     debug = optparse.make_option(
         '', '--debug', action='store', metavar="OPTS",
-        help="Debug options, separated by commas",
+        help="Debug options, separated by commas. [env: COVERAGE_DEBUG]",
     )
     directory = optparse.make_option(
         '-d', '--directory', action='store', metavar="DIR",
@@ -66,14 +87,14 @@ class Opts(object):
         '', '--include', action='store',
         metavar="PAT1,PAT2,...",
         help=(
-            "Include only files whose paths match one of these patterns. "
+            "Include only files whose paths match one of these patterns. " +
             "Accepts shell-style wildcards, which must be quoted."
         ),
     )
     pylib = optparse.make_option(
         '-L', '--pylib', action='store_true',
         help=(
-            "Measure coverage even inside the Python installed library, "
+            "Measure coverage even inside the Python installed library, " +
             "which isn't done by default."
         ),
     )
@@ -81,15 +102,18 @@ class Opts(object):
         '-m', '--show-missing', action='store_true',
         help="Show line numbers of statements in each module that weren't executed.",
     )
-    skip_covered = optparse.make_option(
-        '--skip-covered', action='store_true',
-        help="Skip files with 100% coverage.",
+    module = optparse.make_option(
+        '-m', '--module', action='store_true',
+        help=(
+            "<pyfile> is an importable Python module, not a script path, " +
+            "to be run as 'python -m' would run it."
+        ),
     )
     omit = optparse.make_option(
         '', '--omit', action='store',
         metavar="PAT1,PAT2,...",
         help=(
-            "Omit files whose paths match one of these patterns. "
+            "Omit files whose paths match one of these patterns. " +
             "Accepts shell-style wildcards, which must be quoted."
         ),
     )
@@ -98,33 +122,71 @@ class Opts(object):
         metavar="OUTFILE",
         help="Write the XML report to this file. Defaults to 'coverage.xml'",
     )
+    output_json = optparse.make_option(
+        '-o', '', action='store', dest="outfile",
+        metavar="OUTFILE",
+        help="Write the JSON report to this file. Defaults to 'coverage.json'",
+    )
+    json_pretty_print = optparse.make_option(
+        '', '--pretty-print', action='store_true',
+        help="Format the JSON for human readers.",
+    )
     parallel_mode = optparse.make_option(
         '-p', '--parallel-mode', action='store_true',
         help=(
-            "Append the machine name, process id and random number to the "
-            ".coverage data file name to simplify collecting data from "
+            "Append the machine name, process id and random number to the " +
+            ".coverage data file name to simplify collecting data from " +
             "many processes."
         ),
     )
-    module = optparse.make_option(
-        '-m', '--module', action='store_true',
+    precision = optparse.make_option(
+        '', '--precision', action='store', metavar='N', type=int,
         help=(
-            "<pyfile> is an importable Python module, not a script path, "
-            "to be run as 'python -m' would run it."
+            "Number of digits after the decimal point to display for " +
+            "reported coverage percentages."
         ),
     )
+    quiet = optparse.make_option(
+        '-q', '--quiet', action='store_true',
+        help="Don't print messages about what is happening.",
+    )
     rcfile = optparse.make_option(
         '', '--rcfile', action='store',
-        help="Specify configuration file.  Defaults to '.coveragerc'",
+        help=(
+            "Specify configuration file. " +
+            "By default '.coveragerc', 'setup.cfg', 'tox.ini', and " +
+            "'pyproject.toml' are tried. [env: COVERAGE_RCFILE]"
+        ),
+    )
+    show_contexts = optparse.make_option(
+        '--show-contexts', action='store_true',
+        help="Show contexts for covered lines.",
+    )
+    skip_covered = optparse.make_option(
+        '--skip-covered', action='store_true',
+        help="Skip files with 100% coverage.",
+    )
+    no_skip_covered = optparse.make_option(
+        '--no-skip-covered', action='store_false', dest='skip_covered',
+        help="Disable --skip-covered.",
+    )
+    skip_empty = optparse.make_option(
+        '--skip-empty', action='store_true',
+        help="Skip files with no code.",
+    )
+    sort = optparse.make_option(
+        '--sort', action='store', metavar='COLUMN',
+        help="Sort the report by the named column: name, stmts, miss, branch, brpart, or cover. " +
+             "Default is name."
     )
     source = optparse.make_option(
         '', '--source', action='store', metavar="SRC1,SRC2,...",
-        help="A list of packages or directories of code to be measured.",
+        help="A list of directories or importable names of code to measure.",
     )
     timid = optparse.make_option(
         '', '--timid', action='store_true',
         help=(
-            "Use a simpler but slower trace method.  Try this if you get "
+            "Use a simpler but slower trace method. Try this if you get " +
             "seemingly impossible results!"
         ),
     )
@@ -138,7 +200,7 @@ class Opts(object):
     )
 
 
-class CoverageOptionParser(optparse.OptionParser, object):
+class CoverageOptionParser(optparse.OptionParser):
     """Base OptionParser for coverage.py.
 
     Problems don't exit the program.
@@ -147,7 +209,7 @@ class CoverageOptionParser(optparse.OptionParser, object):
     """
 
     def __init__(self, *args, **kwargs):
-        super(CoverageOptionParser, self).__init__(
+        super().__init__(
             add_help_option=False, *args, **kwargs
             )
         self.set_defaults(
@@ -155,19 +217,27 @@ class CoverageOptionParser(optparse.OptionParser, object):
             append=None,
             branch=None,
             concurrency=None,
+            context=None,
+            contexts=None,
             debug=None,
             directory=None,
             fail_under=None,
             help=None,
             ignore_errors=None,
             include=None,
+            keep=None,
             module=None,
             omit=None,
             parallel_mode=None,
+            precision=None,
             pylib=None,
+            quiet=None,
             rcfile=True,
+            show_contexts=None,
             show_missing=None,
             skip_covered=None,
+            skip_empty=None,
+            sort=None,
             source=None,
             timid=None,
             title=None,
@@ -175,11 +245,6 @@ class CoverageOptionParser(optparse.OptionParser, object):
             )
 
         self.disable_interspersed_args()
-        self.help_fn = self.help_noop
-
-    def help_noop(self, error=None, topic=None, parser=None):
-        """No-op help function."""
-        pass
 
     class OptionParserError(Exception):
         """Used to stop the optparse error handler ending the process."""
@@ -192,15 +257,14 @@ class CoverageOptionParser(optparse.OptionParser, object):
 
         """
         try:
-            options, args = \
-                super(CoverageOptionParser, self).parse_args(args, options)
+            options, args = super().parse_args(args, options)
         except self.OptionParserError:
             return False, None, None
         return True, options, args
 
     def error(self, msg):
         """Override optparse.error so sys.exit doesn't get called."""
-        self.help_fn(msg)
+        show_help(msg)
         raise self.OptionParserError
 
 
@@ -208,7 +272,7 @@ class GlobalOptionParser(CoverageOptionParser):
     """Command-line parser for coverage.py global option arguments."""
 
     def __init__(self):
-        super(GlobalOptionParser, self).__init__()
+        super().__init__()
 
         self.add_options([
             Opts.help,
@@ -231,7 +295,7 @@ class CmdOptionParser(CoverageOptionParser):
         """
         if usage:
             usage = "%prog " + usage
-        super(CmdOptionParser, self).__init__(
+        super().__init__(
             usage=usage,
             description=description,
         )
@@ -242,16 +306,16 @@ class CmdOptionParser(CoverageOptionParser):
     def __eq__(self, other):
         # A convenience equality, so that I can put strings in unit test
         # results, and they will compare equal to objects.
-        return (other == "<CmdOptionParser:%s>" % self.cmd)
+        return (other == f"<CmdOptionParser:{self.cmd}>")
 
     __hash__ = None     # This object doesn't need to be hashed.
 
     def get_prog_name(self):
         """Override of an undocumented function in optparse.OptionParser."""
-        program_name = super(CmdOptionParser, self).get_prog_name()
+        program_name = super().get_prog_name()
 
         # Include the sub-command for this parser as part of the command.
-        return "{command} {subcommand}".format(command=program_name, subcommand=self.cmd)
+        return f"{program_name} {self.cmd}"
 
 
 GLOBAL_ARGS = [
@@ -271,7 +335,7 @@ CMDS = {
             ] + GLOBAL_ARGS,
         usage="[options] [modules]",
         description=(
-            "Make annotated copies of the given files, marking statements that are executed "
+            "Make annotated copies of the given files, marking statements that are executed " +
             "with > and statements that are missed with !."
         ),
     ),
@@ -280,14 +344,16 @@ CMDS = {
         "combine",
         [
             Opts.append,
+            Opts.keep,
+            Opts.quiet,
             ] + GLOBAL_ARGS,
         usage="[options] <path1> <path2> ... <pathN>",
         description=(
-            "Combine data from multiple coverage files collected "
-            "with 'run -p'.  The combined results are written to a single "
-            "file representing the union of the data. The positional "
-            "arguments are data files or directories containing data files. "
-            "If no paths are provided, data files in the default data file's "
+            "Combine data from multiple coverage files collected " +
+            "with 'run -p'.  The combined results are written to a single " +
+            "file representing the union of the data. The positional " +
+            "arguments are data files or directories containing data files. " +
+            "If no paths are provided, data files in the default data file's " +
             "directory are combined."
         ),
     ),
@@ -296,10 +362,13 @@ CMDS = {
         "debug", GLOBAL_ARGS,
         usage="<topic>",
         description=(
-            "Display information on the internals of coverage.py, "
-            "for diagnosing problems. "
-            "Topics are 'data' to show a summary of the collected data, "
-            "or 'sys' to show installation information."
+            "Display information about the internals of coverage.py, " +
+            "for diagnosing problems. " +
+            "Topics are: " +
+                "'data' to show a summary of the collected data; " +
+                "'sys' to show installation information; " +
+                "'config' to show the configuration; " +
+                "'premain' to show what is calling coverage."
         ),
     ),
 
@@ -317,31 +386,59 @@ CMDS = {
     'html': CmdOptionParser(
         "html",
         [
+            Opts.contexts,
             Opts.directory,
             Opts.fail_under,
             Opts.ignore_errors,
             Opts.include,
             Opts.omit,
-            Opts.title,
+            Opts.precision,
+            Opts.quiet,
+            Opts.show_contexts,
             Opts.skip_covered,
+            Opts.no_skip_covered,
+            Opts.skip_empty,
+            Opts.title,
             ] + GLOBAL_ARGS,
         usage="[options] [modules]",
         description=(
-            "Create an HTML report of the coverage of the files.  "
-            "Each file gets its own page, with the source decorated to show "
+            "Create an HTML report of the coverage of the files.  " +
+            "Each file gets its own page, with the source decorated to show " +
             "executed, excluded, and missed lines."
         ),
     ),
 
+    'json': CmdOptionParser(
+        "json",
+        [
+            Opts.contexts,
+            Opts.fail_under,
+            Opts.ignore_errors,
+            Opts.include,
+            Opts.omit,
+            Opts.output_json,
+            Opts.json_pretty_print,
+            Opts.quiet,
+            Opts.show_contexts,
+            ] + GLOBAL_ARGS,
+        usage="[options] [modules]",
+        description="Generate a JSON report of coverage results."
+    ),
+
     'report': CmdOptionParser(
         "report",
         [
+            Opts.contexts,
             Opts.fail_under,
             Opts.ignore_errors,
             Opts.include,
             Opts.omit,
+            Opts.precision,
+            Opts.sort,
             Opts.show_missing,
             Opts.skip_covered,
+            Opts.no_skip_covered,
+            Opts.skip_empty,
             ] + GLOBAL_ARGS,
         usage="[options] [modules]",
         description="Report coverage statistics on modules."
@@ -353,6 +450,7 @@ CMDS = {
             Opts.append,
             Opts.branch,
             Opts.concurrency,
+            Opts.context,
             Opts.include,
             Opts.module,
             Opts.omit,
@@ -373,6 +471,8 @@ CMDS = {
             Opts.include,
             Opts.omit,
             Opts.output_xml,
+            Opts.quiet,
+            Opts.skip_empty,
             ] + GLOBAL_ARGS,
         usage="[options] [modules]",
         description="Generate an XML report of coverage results."
@@ -380,45 +480,57 @@ CMDS = {
 }
 
 
+def show_help(error=None, topic=None, parser=None):
+    """Display an error message, or the named topic."""
+    assert error or topic or parser
+
+    program_path = sys.argv[0]
+    if program_path.endswith(os.path.sep + '__main__.py'):
+        # The path is the main module of a package; get that path instead.
+        program_path = os.path.dirname(program_path)
+    program_name = os.path.basename(program_path)
+    if env.WINDOWS:
+        # entry_points={'console_scripts':...} on Windows makes files
+        # called coverage.exe, coverage3.exe, and coverage-3.5.exe. These
+        # invoke coverage-script.py, coverage3-script.py, and
+        # coverage-3.5-script.py.  argv[0] is the .py file, but we want to
+        # get back to the original form.
+        auto_suffix = "-script.py"
+        if program_name.endswith(auto_suffix):
+            program_name = program_name[:-len(auto_suffix)]
+
+    help_params = dict(coverage.__dict__)
+    help_params['program_name'] = program_name
+    if CTracer is not None:
+        help_params['extension_modifier'] = 'with C extension'
+    else:
+        help_params['extension_modifier'] = 'without C extension'
+
+    if error:
+        print(error, file=sys.stderr)
+        print(f"Use '{program_name} help' for help.", file=sys.stderr)
+    elif parser:
+        print(parser.format_help().strip())
+        print()
+    else:
+        help_msg = textwrap.dedent(HELP_TOPICS.get(topic, '')).strip()
+        if help_msg:
+            print(help_msg.format(**help_params))
+        else:
+            print(f"Don't know topic {topic!r}")
+    print("Full documentation is at {__url__}".format(**help_params))
+
+
 OK, ERR, FAIL_UNDER = 0, 1, 2
 
 
-class CoverageScript(object):
+class CoverageScript:
     """The command-line interface to coverage.py."""
 
-    def __init__(self, _covpkg=None, _run_python_file=None,
-                 _run_python_module=None, _help_fn=None, _path_exists=None):
-        # _covpkg is for dependency injection, so we can test this code.
-        if _covpkg:
-            self.covpkg = _covpkg
-        else:
-            import coverage
-            self.covpkg = coverage
-
-        # For dependency injection:
-        self.run_python_file = _run_python_file or run_python_file
-        self.run_python_module = _run_python_module or run_python_module
-        self.help_fn = _help_fn or self.help
-        self.path_exists = _path_exists or os.path.exists
+    def __init__(self):
         self.global_option = False
-
         self.coverage = None
 
-        program_path = sys.argv[0]
-        if program_path.endswith(os.path.sep + '__main__.py'):
-            # The path is the main module of a package; get that path instead.
-            program_path = os.path.dirname(program_path)
-        self.program_name = os.path.basename(program_path)
-        if env.WINDOWS:
-            # entry_points={'console_scripts':...} on Windows makes files
-            # called coverage.exe, coverage3.exe, and coverage-3.5.exe. These
-            # invoke coverage-script.py, coverage3-script.py, and
-            # coverage-3.5-script.py.  argv[0] is the .py file, but we want to
-            # get back to the original form.
-            auto_suffix = "-script.py"
-            if self.program_name.endswith(auto_suffix):
-                self.program_name = self.program_name[:-len(auto_suffix)]
-
     def command_line(self, argv):
         """The bulk of the command line interface to coverage.py.
 
@@ -429,7 +541,7 @@ class CoverageScript(object):
         """
         # Collect the command-line options.
         if not argv:
-            self.help_fn(topic='minimum_help')
+            show_help(topic='minimum_help')
             return OK
 
         # The command syntax we parse depends on the first argument.  Global
@@ -440,11 +552,10 @@ class CoverageScript(object):
         else:
             parser = CMDS.get(argv[0])
             if not parser:
-                self.help_fn("Unknown command: '%s'" % argv[0])
+                show_help(f"Unknown command: {argv[0]!r}")
                 return ERR
             argv = argv[1:]
 
-        parser.help_fn = self.help_fn
         ok, options, args = parser.parse_args_ok(argv)
         if not ok:
             return ERR
@@ -453,18 +564,15 @@ class CoverageScript(object):
         if self.do_help(options, args, parser):
             return OK
 
-        # We need to be able to import from the current directory, because
-        # plugins may try to, for example, to read Django settings.
-        sys.path[0] = ''
-
         # Listify the list options.
         source = unshell_list(options.source)
         omit = unshell_list(options.omit)
         include = unshell_list(options.include)
         debug = unshell_list(options.debug)
+        contexts = unshell_list(options.contexts)
 
         # Do something.
-        self.coverage = self.covpkg.Coverage(
+        self.coverage = Coverage(
             data_suffix=options.parallel_mode,
             cover_pylib=options.pylib,
             timid=options.timid,
@@ -475,6 +583,9 @@ class CoverageScript(object):
             include=include,
             debug=debug,
             concurrency=options.concurrency,
+            check_preimported=True,
+            context=options.context,
+            messages=not options.quiet,
             )
 
         if options.action == "debug":
@@ -491,7 +602,7 @@ class CoverageScript(object):
             if options.append:
                 self.coverage.load()
             data_dirs = args or None
-            self.coverage.combine(data_dirs, strict=True)
+            self.coverage.combine(data_dirs, strict=True, keep=bool(options.keep))
             self.coverage.save()
             return OK
 
@@ -501,25 +612,54 @@ class CoverageScript(object):
             ignore_errors=options.ignore_errors,
             omit=omit,
             include=include,
+            contexts=contexts,
             )
 
+        # We need to be able to import from the current directory, because
+        # plugins may try to, for example, to read Django settings.
+        sys.path.insert(0, '')
+
         self.coverage.load()
 
         total = None
         if options.action == "report":
             total = self.coverage.report(
                 show_missing=options.show_missing,
-                skip_covered=options.skip_covered, **report_args)
+                skip_covered=options.skip_covered,
+                skip_empty=options.skip_empty,
+                precision=options.precision,
+                sort=options.sort,
+                **report_args
+                )
         elif options.action == "annotate":
-            self.coverage.annotate(
-                directory=options.directory, **report_args)
+            self.coverage.annotate(directory=options.directory, **report_args)
         elif options.action == "html":
             total = self.coverage.html_report(
-                directory=options.directory, title=options.title,
-                skip_covered=options.skip_covered, **report_args)
+                directory=options.directory,
+                title=options.title,
+                skip_covered=options.skip_covered,
+                skip_empty=options.skip_empty,
+                show_contexts=options.show_contexts,
+                precision=options.precision,
+                **report_args
+                )
         elif options.action == "xml":
             outfile = options.outfile
-            total = self.coverage.xml_report(outfile=outfile, **report_args)
+            total = self.coverage.xml_report(
+                outfile=outfile, skip_empty=options.skip_empty,
+                **report_args
+                )
+        elif options.action == "json":
+            outfile = options.outfile
+            total = self.coverage.json_report(
+                outfile=outfile,
+                pretty_print=options.pretty_print,
+                show_contexts=options.show_contexts,
+                **report_args
+            )
+        else:
+            # There are no other possible actions.
+            raise AssertionError
 
         if total is not None:
             # Apply the command line fail-under options, and then use the config
@@ -530,31 +670,16 @@ class CoverageScript(object):
             fail_under = self.coverage.get_option("report:fail_under")
             precision = self.coverage.get_option("report:precision")
             if should_fail_under(total, fail_under, precision):
+                msg = "total of {total} is less than fail-under={fail_under:.{p}f}".format(
+                    total=Numbers(precision=precision).display_covered(total),
+                    fail_under=fail_under,
+                    p=precision,
+                )
+                print("Coverage failure:", msg)
                 return FAIL_UNDER
 
         return OK
 
-    def help(self, error=None, topic=None, parser=None):
-        """Display an error message, or the named topic."""
-        assert error or topic or parser
-        if error:
-            print(error, file=sys.stderr)
-            print("Use '%s help' for help." % (self.program_name,), file=sys.stderr)
-        elif parser:
-            print(parser.format_help().strip())
-        else:
-            help_params = dict(self.covpkg.__dict__)
-            help_params['program_name'] = self.program_name
-            if CTracer is not None:
-                help_params['extension_modifier'] = 'with C extension'
-            else:
-                help_params['extension_modifier'] = 'without C extension'
-            help_msg = textwrap.dedent(HELP_TOPICS.get(topic, '')).strip()
-            if help_msg:
-                print(help_msg.format(**help_params))
-            else:
-                print("Don't know topic %r" % topic)
-
     def do_help(self, options, args, parser):
         """Deal with help requests.
 
@@ -564,9 +689,9 @@ class CoverageScript(object):
         # Handle help.
         if options.help:
             if self.global_option:
-                self.help_fn(topic='help')
+                show_help(topic='help')
             else:
-                self.help_fn(parser=parser)
+                show_help(parser=parser)
             return True
 
         if options.action == "help":
@@ -574,16 +699,16 @@ class CoverageScript(object):
                 for a in args:
                     parser = CMDS.get(a)
                     if parser:
-                        self.help_fn(parser=parser)
+                        show_help(parser=parser)
                     else:
-                        self.help_fn(topic=a)
+                        show_help(topic=a)
             else:
-                self.help_fn(topic='help')
+                show_help(topic='help')
             return True
 
         # Handle version.
         if options.version:
-            self.help_fn(topic='version')
+            show_help(topic='version')
             return True
 
         return False
@@ -592,11 +717,22 @@ class CoverageScript(object):
         """Implementation of 'coverage run'."""
 
         if not args:
-            self.help_fn("Nothing to do.")
+            if options.module:
+                # Specified -m with nothing else.
+                show_help("No module specified for -m")
+                return ERR
+            command_line = self.coverage.get_option("run:command_line")
+            if command_line is not None:
+                args = shlex.split(command_line)
+                if args and args[0] in {"-m", "--module"}:
+                    options.module = True
+                    args = args[1:]
+        if not args:
+            show_help("Nothing to do.")
             return ERR
 
         if options.append and self.coverage.get_option("run:parallel"):
-            self.help_fn("Can't append to data files in parallel mode.")
+            show_help("Can't append to data files in parallel mode.")
             return ERR
 
         if options.concurrency == "multiprocessing":
@@ -606,35 +742,32 @@ class CoverageScript(object):
                 # As it happens, all of these options have no default, meaning
                 # they will be None if they have not been specified.
                 if getattr(options, opt_name) is not None:
-                    self.help_fn(
-                        "Options affecting multiprocessing must be specified "
-                        "in a configuration file."
+                    show_help(
+                        "Options affecting multiprocessing must only be specified " +
+                        "in a configuration file.\n" +
+                        f"Remove --{opt_name} from the command line."
                     )
                     return ERR
 
-        if not self.coverage.get_option("run:parallel"):
-            if not options.append:
-                self.coverage.erase()
+        os.environ["COVERAGE_RUN"] = "true"
+
+        runner = PyRunner(args, as_module=bool(options.module))
+        runner.prepare()
+
+        if options.append:
+            self.coverage.load()
 
         # Run the script.
         self.coverage.start()
         code_ran = True
         try:
-            if options.module:
-                self.run_python_module(args[0], args)
-            else:
-                filename = args[0]
-                self.run_python_file(filename, args)
+            runner.run()
         except NoSource:
             code_ran = False
             raise
         finally:
             self.coverage.stop()
             if code_ran:
-                if options.append:
-                    data_file = self.coverage.get_option("run:data_file")
-                    if self.path_exists(data_file):
-                        self.coverage.combine(data_paths=[data_file])
                 self.coverage.save()
 
         return OK
@@ -643,7 +776,7 @@ class CoverageScript(object):
         """Implementation of 'coverage debug'."""
 
         if not args:
-            self.help_fn("What information would you like: config, data, sys?")
+            show_help("What information would you like: config, data, sys, premain?")
             return ERR
 
         for info in args:
@@ -651,22 +784,22 @@ class CoverageScript(object):
                 sys_info = self.coverage.sys_info()
                 print(info_header("sys"))
                 for line in info_formatter(sys_info):
-                    print(" %s" % line)
+                    print(f" {line}")
             elif info == 'data':
                 self.coverage.load()
-                data = self.coverage.data
+                data = self.coverage.get_data()
                 print(info_header("data"))
-                print("path: %s" % self.coverage.data_files.filename)
+                print(f"path: {data.data_filename()}")
                 if data:
-                    print("has_arcs: %r" % data.has_arcs())
-                    summary = data.line_counts(fullpath=True)
-                    filenames = sorted(summary.keys())
-                    print("\n%d files:" % len(filenames))
+                    print(f"has_arcs: {data.has_arcs()!r}")
+                    summary = line_counts(data, fullpath=True)
+                    filenames = human_sorted(summary.keys())
+                    print(f"\n{len(filenames)} files:")
                     for f in filenames:
-                        line = "%s: %d lines" % (f, summary[f])
+                        line = f"{f}: {summary[f]} lines"
                         plugin = data.file_tracer(f)
                         if plugin:
-                            line += " [%s]" % plugin
+                            line += f" [{plugin}]"
                         print(line)
                 else:
                     print("No data collected")
@@ -674,9 +807,12 @@ class CoverageScript(object):
                 print(info_header("config"))
                 config_info = self.coverage.config.__dict__.items()
                 for line in info_formatter(config_info):
-                    print(" %s" % line)
+                    print(f" {line}")
+            elif info == "premain":
+                print(info_header("premain"))
+                print(short_stack())
             else:
-                self.help_fn("Don't know what you mean by %r" % info)
+                show_help(f"Don't know what you mean by {info!r}")
                 return ERR
 
         return OK
@@ -719,24 +855,24 @@ HELP_TOPICS = {
         Commands:
             annotate    Annotate source files with execution information.
             combine     Combine a number of data files.
+            debug       Display information about the internals of coverage.py
             erase       Erase previously collected coverage data.
             help        Get help on using coverage.py.
             html        Create an HTML report.
+            json        Create a JSON report of coverage results.
             report      Report coverage stats on modules.
             run         Run a Python program and measure code execution.
             xml         Create an XML report of coverage results.
 
         Use "{program_name} help <command>" for detailed help on any command.
-        For full documentation, see {__url__}
     """,
 
     'minimum_help': """\
-        Code coverage for Python.  Use '{program_name} help' for help.
+        Code coverage for Python, version {__version__} {extension_modifier}.  Use '{program_name} help' for help.
     """,
 
     'version': """\
         Coverage.py, version {__version__} {extension_modifier}
-        Documentation at {__url__}
     """,
 }
 
@@ -759,7 +895,8 @@ def main(argv=None):
         status = ERR
     except BaseCoverageException as err:
         # A controlled error inside coverage.py: print the message to the user.
-        print(err)
+        msg = err.args[0]
+        print(msg)
         status = ERR
     except SystemExit as err:
         # The user called `sys.exit()`.  Exit with their argument, if any.
@@ -768,3 +905,22 @@ def main(argv=None):
         else:
             status = None
     return status
+
+# Profiling using ox_profile.  Install it from GitHub:
+#   pip install git+https://github.com/emin63/ox_profile.git
+#
+# $set_env.py: COVERAGE_PROFILE - Set to use ox_profile.
+_profile = os.environ.get("COVERAGE_PROFILE", "")
+if _profile:                                                # pragma: debugging
+    from ox_profile.core.launchers import SimpleLauncher    # pylint: disable=import-error
+    original_main = main
+
+    def main(argv=None):                                    # pylint: disable=function-redefined
+        """A wrapper around main that profiles."""
+        profiler = SimpleLauncher.launch()
+        try:
+            return original_main(argv)
+        finally:
+            data, _ = profiler.query(re_filter='coverage', max_records=100)
+            print(profiler.show(query=data, limit=100, sep='', col=''))
+            profiler.cancel()
